@@ -24,6 +24,7 @@
 // 			conn, err := ln.AcceptSocks()
 // 			if err != nil {
 // 				if e, ok := err.(net.Error); ok && e.Temporary() {
+// 					pt.Log(pt.LogSeverityError, "accept error: " + err.Error())
 // 					continue
 // 				}
 // 				return err
@@ -42,7 +43,7 @@
 // 		if ptInfo.ProxyURL != nil {
 // 			// you need to interpret the proxy URL yourself
 // 			// call pt.ProxyDone instead if it's a type you understand
-// 			pt.ProxyError("proxy %s is not supported")
+// 			pt.ProxyError(fmt.Sprintf("proxy %s is not supported", ptInfo.ProxyURL))
 // 			os.Exit(1)
 // 		}
 // 		for _, methodName := range ptInfo.MethodNames {
@@ -83,6 +84,7 @@
 // 				if e, ok := err.(net.Error); ok && e.Temporary() {
 // 					continue
 // 				}
+// 				pt.Log(pt.LogSeverityError, "accept error: " + err.Error())
 // 				return err
 // 			}
 // 			go handler(conn)
@@ -113,17 +115,17 @@
 // 		pt.SmethodsDone()
 // 	}
 //
-// Some additional care is needed to handle SIGINT and shutdown properly. See
+// Some additional care is needed to handle signals and shutdown properly. See
 // the example programs dummy-client and dummy-server.
 //
 // Tor pluggable transports specification:
 // https://spec.torproject.org/pt-spec
 //
 // Extended ORPort:
-// https://gitweb.torproject.org/torspec.git/tree/proposals/196-transport-control-ports.txt.
+// https://gitweb.torproject.org/torspec.git/tree/proposals/196-transport-control-ports.txt
 //
 // Extended ORPort Authentication:
-// https://gitweb.torproject.org/torspec.git/tree/proposals/217-ext-orport-auth.txt.
+// https://gitweb.torproject.org/torspec.git/tree/proposals/217-ext-orport-auth.txt
 //
 // Pluggable Transport through SOCKS proxy:
 // https://gitweb.torproject.org/torspec.git/tree/proposals/232-pluggable-transports-through-proxy.txt
@@ -217,19 +219,18 @@ func getenvRequired(key string) (string, error) {
 // <KeywordChar> ::= <any US-ASCII alphanumeric, dash, and underscore>
 func keywordIsSafe(keyword string) bool {
 	for _, b := range []byte(keyword) {
-		if b >= '0' && b <= '9' {
+		switch {
+		case '0' <= b && b <= '9':
 			continue
-		}
-		if b >= 'A' && b <= 'Z' {
+		case 'A' <= b && b <= 'Z':
 			continue
-		}
-		if b >= 'a' && b <= 'z' {
+		case 'a' <= b && b <= 'z':
 			continue
-		}
-		if b == '-' || b == '_' {
+		case b == '-' || b == '_':
 			continue
+		default:
+			return false
 		}
-		return false
 	}
 	return true
 }
@@ -300,7 +301,7 @@ func SmethodError(methodName, msg string) error {
 // Emit a PROXY-ERROR line with explanation text. Returns a representation of
 // the error.
 func ProxyError(msg string) error {
-	return doError("PROXY-ERROR %s\n", msg)
+	return doError("PROXY-ERROR", msg)
 }
 
 // Emit a CMETHOD line. socks must be "socks4" or "socks5". Call this once for
@@ -348,6 +349,69 @@ func ProxyDone() {
 	fmt.Fprintf(Stdout, "PROXY DONE\n")
 }
 
+// Unexported type to represent log severities, preventing external callers from
+// inventing new severity strings that may violate quoting rules.
+//
+// pt-spec.txt section 3.3.4 specifies quoting for MESSAGE, but not for
+// SEVERITY, and the example shows an unquoted "SEVERITY=debug". While we know
+// tor's parser permits quoting of SEVERITY, it's not actually specified.
+// Therefore we we need to guard against callers passing a string that violates
+// the global protocol constraint of "any US-ASCII character but NUL or NL." So
+// here, we instantiate exactly the five supported severities, using a type that
+// cannot be constructed outside the package.
+type logSeverity struct {
+	string
+}
+
+// Severity levels for the Log function.
+var (
+	LogSeverityError   = logSeverity{"error"}
+	LogSeverityWarning = logSeverity{"warning"}
+	LogSeverityNotice  = logSeverity{"notice"}
+	LogSeverityInfo    = logSeverity{"info"}
+	LogSeverityDebug   = logSeverity{"debug"}
+)
+
+// Encode a string according to the CString rules of section 2.1.1 in
+// control-spec.txt.
+// 	CString = DQUOTE *qcontent DQUOTE
+// "...in a CString, the escapes '\n', '\t', '\r', and the octal escapes '\0'
+// ... '\377' represent newline, tab, carriage return, and the 256 possible
+// octet values respectively."
+// RFC 2822 section 3.2.5 in turn defines what byte values we need to escape:
+// everything but
+// 	NO-WS-CTL /     ; Non white space controls
+// 	%d33 /          ; The rest of the US-ASCII
+// 	%d35-91 /       ;  characters not including "\"
+// 	%d93-126        ;  or the quote character
+// Technically control-spec.txt requires us to escape the space character (32),
+// but it is an error in the spec: https://bugs.torproject.org/29432.
+//
+// We additionally need to ensure that whatever we return passes argIsSafe,
+// because strings encoded by this function are printed verbatim by Log.
+func encodeCString(s string) string {
+	result := bytes.NewBuffer([]byte{})
+	result.WriteByte('"')
+	for _, c := range []byte(s) {
+		if c == 32 || c == 33 || (35 <= c && c <= 91) || (93 <= c && c <= 126) {
+			result.WriteByte(c)
+		} else {
+			fmt.Fprintf(result, "\\%03o", c)
+		}
+	}
+	result.WriteByte('"')
+	return result.String()
+}
+
+// Emit a LOG message with the given severity (one of LogSeverityError,
+// LogSeverityWarning, LogSeverityNotice, LogSeverityInfo, or LogSeverityDebug).
+func Log(severity logSeverity, message string) {
+	// "<Message> contains the log message which can be a String or CString..."
+	// encodeCString always makes the string safe to emit; i.e., it
+	// satisfies argIsSafe.
+	line("LOG", "SEVERITY="+severity.string, "MESSAGE="+encodeCString(message))
+}
+
 // Get a pluggable transports version offered by Tor and understood by us, if
 // any. The only version we understand is "1". This function reads the
 // environment variable TOR_PT_MANAGED_TRANSPORT_VER.
@@ -449,7 +513,7 @@ type ClientInfo struct {
 // was a list of transport names to use in case Tor requested "*". That feature
 // was never implemented and has been removed from the pluggable transports
 // specification.
-// https://trac.torproject.org/projects/tor/ticket/15612
+// https://bugs.torproject.org/15612
 func ClientSetup(_ []string) (info ClientInfo, err error) {
 	ver, err := getManagedTransportVer()
 	if err != nil {
@@ -494,6 +558,7 @@ func resolveAddr(addrStr string) (*net.TCPAddr, error) {
 		// Before the fixing of bug #7011, tor doesn't put brackets around IPv6
 		// addresses. Split after the last colon, assuming it is a port
 		// separator, and try adding the brackets.
+		// https://bugs.torproject.org/7011
 		parts := strings.Split(addrStr, ":")
 		if len(parts) <= 2 {
 			return nil, err
@@ -556,6 +621,7 @@ func getServerBindaddrs() ([]Bindaddr, error) {
 	if err != nil {
 		return nil, err
 	}
+	seenMethods := make(map[string]bool)
 	for _, spec := range strings.Split(serverBindaddr, ",") {
 		var bindaddr Bindaddr
 
@@ -564,6 +630,12 @@ func getServerBindaddrs() ([]Bindaddr, error) {
 			return nil, envError(fmt.Sprintf("TOR_PT_SERVER_BINDADDR: %q: doesn't contain \"-\"", spec))
 		}
 		bindaddr.MethodName = parts[0]
+		// Check for duplicate method names: "Applications MUST NOT set
+		// more than one <address>:<port> pair per PT name."
+		if seenMethods[bindaddr.MethodName] {
+			return nil, envError(fmt.Sprintf("TOR_PT_SERVER_BINDADDR: %q: duplicate method name %q", spec, bindaddr.MethodName))
+		}
+		seenMethods[bindaddr.MethodName] = true
 		addr, err := resolveAddr(parts[1])
 		if err != nil {
 			return nil, envError(fmt.Sprintf("TOR_PT_SERVER_BINDADDR: %q: %s", spec, err.Error()))
@@ -608,13 +680,18 @@ func readAuthCookie(f io.Reader) ([]byte, error) {
 }
 
 // Read and validate the contents of an auth cookie file. Returns the 32-byte
-// cookie. See section 4.2.1.2 of pt-spec.txt.
-func readAuthCookieFile(filename string) ([]byte, error) {
+// cookie. See section 4.2.1.2 of 217-ext-orport-auth.txt.
+func readAuthCookieFile(filename string) (cookie []byte, err error) {
 	f, err := os.Open(filename)
 	if err != nil {
 		return nil, err
 	}
-	defer f.Close()
+	defer func() {
+		closeErr := f.Close()
+		if err == nil {
+			err = closeErr
+		}
+	}()
 
 	return readAuthCookie(f)
 }
@@ -647,7 +724,7 @@ type ServerInfo struct {
 // was a list of transport names to use in case Tor requested "*". That feature
 // was never implemented and has been removed from the pluggable transports
 // specification.
-// https://trac.torproject.org/projects/tor/ticket/15612
+// https://bugs.torproject.org/15612
 func ServerSetup(_ []string) (info ServerInfo, err error) {
 	ver, err := getManagedTransportVer()
 	if err != nil {
@@ -766,6 +843,7 @@ func extOrPortAuthenticate(s io.ReadWriter, info *ServerInfo) error {
 	// Work around tor bug #15240 where the auth cookie is generated after
 	// pluggable transports are launched, leading to a stale cookie getting
 	// cached forever if it is only read once as part of ServerSetup.
+	// https://bugs.torproject.org/15240
 	authCookie, err := readAuthCookieFile(info.AuthCookiePath)
 	if err != nil {
 		return fmt.Errorf("error reading TOR_PT_AUTH_COOKIE_FILE %q: %s", info.AuthCookiePath, err.Error())
@@ -794,7 +872,7 @@ func extOrPortAuthenticate(s io.ReadWriter, info *ServerInfo) error {
 	return nil
 }
 
-// See section 3.1 of 196-transport-control-ports.txt.
+// See section 3.1.1 of 196-transport-control-ports.txt.
 const (
 	extOrCmdDone      = 0x0000
 	extOrCmdUserAddr  = 0x0001
@@ -875,7 +953,7 @@ func extOrPortRecvCommand(s io.Reader) (cmd uint16, body []byte, err error) {
 // OKAY or DENY response command from the server. If addr or methodName is "",
 // the corresponding command is not sent. Returns nil if and only if OKAY is
 // received.
-func extOrPortSetup(s io.ReadWriter, addr, methodName string) error {
+func extOrPortSetMetadata(s io.ReadWriter, addr, methodName string) error {
 	var err error
 
 	if addr != "" {
@@ -907,6 +985,27 @@ func extOrPortSetup(s io.ReadWriter, addr, methodName string) error {
 	return nil
 }
 
+func extOrPortSetup(s net.Conn, timeout time.Duration,
+	info *ServerInfo, addr, methodName string) error {
+	err := s.SetDeadline(time.Now().Add(5 * time.Second))
+	if err != nil {
+		return err
+	}
+	err = extOrPortAuthenticate(s, info)
+	if err != nil {
+		return err
+	}
+	err = extOrPortSetMetadata(s, addr, methodName)
+	if err != nil {
+		return err
+	}
+	err = s.SetDeadline(time.Time{})
+	if err != nil {
+		return err
+	}
+	return nil
+}
+
 // Dial info.ExtendedOrAddr if defined, or else info.OrAddr, and return an open
 // *net.TCPConn. If connecting to the extended OR port, extended OR port
 // authentication à la 217-ext-orport-auth.txt is done before returning; an
@@ -924,18 +1023,11 @@ func DialOr(info *ServerInfo, addr, methodName string) (*net.TCPConn, error) {
 	if err != nil {
 		return nil, err
 	}
-	s.SetDeadline(time.Now().Add(5 * time.Second))
-	err = extOrPortAuthenticate(s, info)
+	err = extOrPortSetup(s, 5*time.Second, info, addr, methodName)
 	if err != nil {
 		s.Close()
 		return nil, err
 	}
-	err = extOrPortSetup(s, addr, methodName)
-	if err != nil {
-		s.Close()
-		return nil, err
-	}
-	s.SetDeadline(time.Time{})
 
 	return s, nil
 }
